/**
 * @name InterTween - Interpolation powers Tween
 * @description The lightweight, fastest, smartest, effecient value interpolator with no-dependecy, zero-configuration and relative interpolation
 * @author dalisoft (https://github.com/dalisoft)
 * @license MIT-License
 */
(function(root, factory) {
        if (typeof define === 'function' && define["amd"]) {
            define([], factory);
        } else if (typeof module !== 'undefined' && module["exports"]) {
            module["exports"] = factory();
        } else {
            root["InterTween"] = factory();
        }
    }
    (typeof(window) !== 'undefined' ? window : this, function() {
        // RegExp variables
        var colorMatch = /rgb/g;
        var isIncrementReqForColor = /argb/g;
        // This RegExp (numRegExp) is original from @jkroso string tweening and optimized by @dalisoft
        var numRegExp =
            /\s+|([A-Za-z?().,{}:""[\]#\%]+)|([-+/*%]+=)?([-+*/%]+)?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g;
        var hexColor = /^#([0-9a-f]{6}|[0-9a-f]{3})$/i;
        var trimRegExp = /\n|\r|\t/g;
        var colorSpaceExp = /, | ,| , /g;
        var rgbMax = 255;
        // Helpers
        function s2f(val) {
            var floatedVal = parseFloat(val);
            return typeof floatedVal === "number" && !isNaN(floatedVal) ? floatedVal : val;
        }
        var isArray = Array.isArray || function(arr) {
            return arr instanceof Array
        };
        var propRegExp = /([.\[])/g;
        var replaceBrace = /\]/g;
        var propExtract = function(obj, property) {
            var value = obj[property];
            var props = property.replace(replaceBrace, '')
                .split(propRegExp);
            var propsLastIndex = props.length - 1;
            var lastArr = Array.isArray(obj);
            var lastObj = typeof obj === 'object' && !lastArr;
            if (lastObj) {
                obj[property] = null;
                delete obj[property];
            } else if (lastArr) {
                obj.splice(property, 1);
            }
            return props.reduce(function(nested, prop, index) {
                if (lastArr) {
                    if (prop !== '.' && prop !== '[') {
                        prop *= 1;
                    }
                }
                var nextProp = props[index + 1];
                var nextIsArray = nextProp === '[';
                if (prop === '.' || prop === '[') {
                    if (prop === '.') {
                        lastObj = true;
                        lastArr = false;
                    } else if (prop === '[') {
                        lastObj = false;
                        lastArr = true;
                    }
                    return nested;
                } else if (nested[prop] === undefined) {
                    if (lastArr || lastObj) {
                        nested[prop] = index === propsLastIndex ? value : lastArr || nextIsArray ? [] :
                            lastObj ? {} : null;
                        lastObj = lastArr = false;
                        return nested[prop];
                    }
                } else if (nested[prop] !== undefined) {
                    if (index === propsLastIndex) {
                        nested[prop] = value;
                    }
                    return nested[prop];
                }
                return nested;
            }, obj);
        };

        var setnested = function(nested) {
            if (typeof nested === 'object' && !!nested) {
                for (var prop in nested) {
                    if (prop.indexOf('.') !== -1 || prop.indexOf('[') !== -1) {
                        propExtract(nested, prop);
                    } else if (typeof nested[prop] === 'object' && !!nested[prop]) {
                        var nested2 = nested[prop];
                        for (var prop2 in nested2) {
                            if (prop2.indexOf('.') !== -1 || prop2.indexOf('[') !== -1) {
                                propExtract(nested2, prop2);
                            } else if (typeof nested2[prop2] === 'object' && !!nested2[prop2]) {
                                var nested3 = nested2[prop2];
                                for (var prop3 in nested3) {
                                    if (prop3.indexOf('.') !== -1 || prop3.indexOf('[') !== -1) {
                                        propExtract(nested3, prop3);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            return nested;
        }

        function h2r_f(all, hex) {
            var r;
            var g;
            var b;
            if (hex.length === 3) {
                r = hex[0];
                g = hex[1];
                b = hex[2];
                hex = r + r + g + g + b + b;
            }
            var color = parseInt(hex, 16);
            r = color >> 16 & rgbMax;
            g = color >> 8 & rgbMax;
            b = color & rgbMax;
            return "rgb(" + r + "," + g + "," + b + ")";
        };

        function trim(str) {
            return typeof str === "string" ? str.replace(trimRegExp, "")
                .replace(colorSpaceExp, ',') : str;
        };
        var relativeModes = {
            '+=': 1,
            '-=': 1,
            '*=': 2,
            '/=': 3,
            '%=': 4
        };

        function r2n(s, e) {
            if (typeof e === 'number') {
                return e;
            } else {
                var rv = relativeModes[e.substr(0, 2)],
                    v = e.substr(2);
                if (rv === 1) {
                    var e2 = e[0] + v;
                    return s + parseFloat(e2);
                } else if (rv === 2) {
                    return s * +v;
                } else if (rv === 3) {
                    return s / +v;
                } else if (rv === 4) {
                    return s * (+v / 100);
                }
            }
            return e;
        }
        var h2r = function(hex) {
            return typeof hex !== 'string' ? hex : trim(hex)
                .replace(hexColor, h2r_f);
        }

        function s2n(str) {
            var hr = h2r(str);
            return typeof hr === 'string' ? hr.match(numRegExp)
                .map(s2f) : str;
        }
        // Splitted functions
        function stringTween(s, e, d, easing) {
            d = d !== null && d !== undefined ? d : 10000;
            if (!numRegExp.test(e)) return e;
            var sv = s2n(s);
            var ev = s2n(e);
            var uv = unitTween(sv, ev, d, easing);
            if (uv) {
                return uv;
            }
            uv = null;
            var cm = null;
            var cmls = null;
            var rv = [];
            var i = 0;
            var len = ev.length;
            for (; i < len; i++) {
                var ve = ev[i],
                    vs = sv[i];
                rv[i] = typeof ve === 'string' && ve.indexOf('=') === 1 ? ve : null;
                if (isIncrementReqForColor.test(ve)) {
                    cm = i + 2;
                    cmls = i + 11;
                } else if (colorMatch.test(ve)) {
                    cm = i;
                    cmls = i + 9;
                }
                ev[i] = vs === ve ? null : rv[i] !== null ? r2n(vs, ve) : ve;
            }
            return function(t, vt, ease) {
                if (!ease) {
                    ease = easing;
                }
                vt = vt !== undefined ? vt : t;
                var str = '';
                i = 0;
                for (; i < len; i++) {
                    var a = sv[i],
                        b = ev[i],
                        r = rv[i],
                        _t = typeof ease === 'function' ? ease(t) : t;
                    str += typeof b === 'number' ? cm !== null && i > cm && i < cmls ? (a + (b - a) * _t) | 0 :
                        (((a + (b - a) * _t) * d) | 0) / d : a;
                    if (vt === 1 && r !== null) {
                        sv[i] = b;
                        ev[i] = r2n(b, r);
                    }
                }
                return str;
            }
        }

        function tweenThemTo(sv, ev, d, easing) {
            var vs = [];
            for (var i = 0, len = sv.length; i < len; i++) {
                var s = sv[i];
                vs[i] = isArray(s) ? arrayTween(s, ev, d, easing && easing[i] || easing) : typeof s === 'object' ?
                    objectTween(s, ev, d, easing && easing[i] || easing) : typeof s ===
                    'string' ? stringTween(s, ev, d, easing && easing[i] || easing) : s;
            }
            return function(t, v, ease) {
				v = v !== undefined ? v : t;
                if (!ease) {
                    ease = easing;
                }
                for (var i = 0, len = vs.length; i < len; i++) {
					var _easing = _easing = typeof(ease) === 'function' ? ease : ease && typeof ease[i] === 'function' ?
                        ease[i] : null;
                    var _t = typeof _easing ? _easing(t) : t;
                    sv[i] = typeof vs[i] === 'function' ? vs[i](_easing ? t : _t, v, ease) : typeof vs[i] === 'number' ? vs[i] + (ev -
                        vs[i]) * _t : vs[i];
                }
                return sv;
            }
        };

        function parseInterpolatables(sv, ev, d, easing) {
            var vs = [];
            for (var i = 0, len = ev.length; i < len; i++) {
                var e = ev[i];
                vs[i] = mainTween(i === 0 ? sv : ev[i - 1], e, d, easing && easing[i] || easing);
            }
            var lastItem = ev[ev.length - 1];
            vs.push(mainTween(lastItem, lastItem, d, easing && easing[ev.length - 1] || easing));
            var endLength = vs.length - 1;
            return function(t) {
                var totalTime = t * endLength;
                var roundedTime = Math.max(0, Math.floor(totalTime));
                var elapsed = totalTime - roundedTime;
                var item = vs[roundedTime];
                var interpolated = typeof item === 'function' ? item(elapsed) : item;
                return interpolated;
            };
        }

        function arrayTween(sv, ev, d, easing) {
            d = d !== null && d !== undefined ? d : 10000;
            setnested(ev)
            setnested(sv)
            var s = sv.slice();
            var rv = [];
            var minLength = Math.min(sv.length, ev.length);
            for (var i = 0; i < minLength; i++) {
                var vs = s[i],
                    ve = ev[i];
				if (vs === undefined || vs === null) {
					if (typeof ve === 'function' || (ve && ve.update)) {
						s[i] = ve;
					}
					continue
				}
                rv[i] = typeof ve === 'string' && ve.indexOf('=') === 1 ? ve : null;
                s[i] = ve.nodeType || ve.update ? ve.update : vs === ve ? null : isArray(ve) ? isArray(vs) && ve.length ===
                    vs.length ? arrayTween(vs, ve, d, easing && easing[i] || easing) : parseInterpolatables(vs, ve,
                        d, easing && easing[i] || easing) : isArray(vs) ? tweenThemTo(vs,
                        ve, d, easing && easing[i] || easing) : typeof vs === 'object' ? objectTween(vs, ve, d,
                        easing && easing[i] || easing) : typeof vs === 'string' ?
                    stringTween(vs, ve, d, easing && easing[i] || easing) : vs !== undefined ? vs : ve;
                ev[i] = rv[i] !== null ? r2n(vs, ve) : ve;
            }
            return function(t, v, ease) {
                if (!ease) {
                    ease = easing;
                }
                v !== undefined ? v : t;
                for (var i = 0; i < minLength; i++) {
                    var a = s[i],
                        b = ev[i],
                        r = rv[i];
                    if (a === null || a === undefined) continue;
					var _easing = _easing = typeof(ease) === 'function' ? ease : ease && typeof ease[i] === 'function' ?
                        ease[i] : null;
                    var _t = _easing ? _easing(t) : t;
                    sv[i] = typeof a === 'number' ? (((a + (b - a) * _t) * d) | 0) / d : typeof a ===
                        'function' ?
                        a(_easing ? t : _t, v, _easing) : a.update ? a.update(_t, v, _easing) : b && b.update ? b.update(_t, v, _easing) : b;
                    if (r && v === 1) {
                        s[i] = b;
                        ev[i] = r2n(s[i], r);
                    }
                }
                return v === 1 ? ev : sv;
            }
        }
        var units = ["px", "pt", "pc", "deg", "rad", "turn", "em", "ex", "cm", "mm", "dm", "inch", "in", "rem",
            "vw", "vh", "vmin", "vmax", "%"
        ];

        function unitTween(sv, ev, d, easing) {
            d = d !== null && d !== undefined ? d : 10000;
            if (ev.length === 2 && sv.length === 2) {
                var unidx = units.indexOf(ev[1]);
                if (unidx !== -1) {
                    var s = +sv[0],
                        e = +ev[0],
                        u = ev[1],
                        r = typeof ev[0] === 'string' && ev[0].indexOf('=') === 1 ? ev[0] : null;
                    if (r) {
                        e = r2n(s, e);
                    }
                    return s === e && ev[0] === u ? ev : function(t, vt, ease) {
                        if (!ease) {
                            ease = easing;
                        }
                        vt !== undefined ? vt : t;
                        var _t = typeof ease === 'function' ? ease(t) : t;
                        var v = ((((s + (e - s) * _t) * d) | 0) / d) + u;
                        if (r && vt === 1) {
                            s = e;
                            e = r2n(s, r);
                        }
                        return v;
                    }
                }
            }
            return false;
        }
		
		var descriptor = Object.getOwnPropertyDescriptor;
        var _getProp = !!descriptor ? descriptor : function() {
            return;
        };

        function objectTween(sv, ev, d, easing) {
            d = d !== null && d !== undefined ? d : 10000;
            var rv = {};
            var s = {};
            var readOnly;
            setnested(ev)
            setnested(sv)
            setnested(easing)
            for (var i in ev) {
                var vs = sv[i],
                    ve = ev[i];
				if (vs === undefined || vs === null) {
					if (typeof ve === 'function' || (ve && ve.update)) {
						s[i] = ve;
					}
					continue
				}
                var propDesc = _getProp(sv, i);
                rv[i] = typeof ve === 'string' && ve.indexOf('=') === 1 ? ve : null;
                if (!!propDesc && (propDesc.writable === false && !propDesc.set)) {
				if (!readOnly) {
					readOnly = {};
				}
                    readOnly[i] = true;
                }
                s[i] = ve.nodeType ? ve : ve.update ? ve : vs === ve ? null : isArray(ve) ? isArray(vs) && ve.length ===
                    vs.length ? arrayTween(vs, ve, d, easing && easing[i] || easing) : parseInterpolatables(vs, ve,
                        d, easing && easing[i] || easing) : isArray(vs) ? tweenThemTo(vs,
                        ve, d, easing && easing[i] || easing) : typeof vs === 'object' ? objectTween(vs, ve, d,
                        easing && easing[i] || easing) : typeof vs === 'string' ?
                    stringTween(vs, ve, d, easing && easing[i] || easing) : vs !== undefined ? vs : ve;
                ev[i] = rv[i] !== null ? r2n(vs, ve) : ve;
            }
            return function(t, vt, ease) {
                if (!ease) {
                    ease = easing;
                }
                vt = vt !== undefined ? vt : t;
                for (var i in s) {
                    var a = s[i],
                        b = ev[i],
                        r = rv[i],
                        _easing = typeof(ease) === 'function' ? ease : ease && typeof ease[i] === 'function' ?
                        ease[i] : null,
                        value = !!_easing ? _easing(t) : t,
                        v = typeof a === 'number' ? (((a + (b - a) * value) * d) | 0) / d : typeof a ===
                        'function' ?
                        a(_easing ? t : value, vt, _easing) : a && a.update ? a.update(value, vt, _easing) : b && b.update ? b.update(value, vt, _easing) :
                        b;
					if (typeof a === 'number') {
					}
                    if (!readOnly || !readOnly[i]) {
                        sv[i] = v;
                    }
                    if (vt === 1 && r) {
                        s[i] = b;
                        ev[i] = r2n(s[i], r);
                    }
                }
                return vt === 1 ? ev : sv;
            }
        }

        function mainTween(sv, ev, d, easing) {
            setnested(easing)
            d = d !== null && d !== undefined ? d : 10000;
            var rv = typeof(ev) === 'string' && typeof sv === 'number' && ev.indexOf('=') === 1 ? ev : null;
            if (rv) {
                ev = r2n(sv, rv);
            }
            return ev.nodeType ? ev : sv.nodeType ? sv : isArray(ev) ? isArray(sv) && sv.length === ev.length ?
                arrayTween(sv, ev, d, easing) : parseInterpolatables(sv, ev, d, easing) : isArray(sv) ? tweenThemTo(
                    sv, ev, d, easing) : typeof ev ===
                'object' ? objectTween(sv, ev, d, easing) : typeof ev === 'string' ? stringTween(sv, ev, d, easing) :
                typeof ev ===
                'function' ? ev : function(t, vt, ease) {
                    vt = vt !== undefined ? vt : t;
                    if (!ease) {
                        ease = easing;
                    }
                    var value = typeof(ease) === 'function' ? ease(t) : t;
                    var vv = typeof ev === 'number' ? (((sv + (ev - sv) * value) * d) | 0) / d : sv
                    if (rv && vt === 1) {
                        sv += ev;
                        ev = r2n(sv, rv);
                    }
                    return vv;
                }
        }
        return mainTween;
    }));